# Copyright (c) QuantCo 2023-2024
# SPDX-License-Identifier: BSD-3-Clause

from __future__ import annotations

import json

import numpy as np
import onnx
import onnxruntime as ort

import ndonnx._data_types as dtypes
from ndonnx._build import (
    _assemble_outputs,
    _deconstruct_inputs,
    _extract_output_names,
    _get_dtype,
)

Dtype = dtypes.StructType | dtypes.CoreType


def run(
    model_proto: onnx.ModelProto, inputs: dict[str, np.ndarray]
) -> dict[str, np.ndarray]:
    # Extract input and output schemas
    input_dtypes, output_dtypes = _get_dtypes(model_proto)
    session = _make_session(model_proto)
    flattened_inputs = _deconstruct_inputs(inputs, input_dtypes)
    output_names = tuple(_extract_output_names(output_dtypes))
    outputs = dict(zip(output_names, session.run(output_names, flattened_inputs)))
    return _assemble_outputs(outputs, output_dtypes)


def _make_session(model_proto: onnx.ModelProto) -> ort.InferenceSession:
    session_options = ort.SessionOptions()
    return ort.InferenceSession(
        model_proto.SerializeToString(), sess_options=session_options
    )


def _get_dtypes(
    model_proto: onnx.ModelProto,
) -> tuple[dict[str, Dtype], dict[str, Dtype]]:
    encoding = json.loads(
        [prop for prop in model_proto.metadata_props if prop.key == "ndonnx_schema"][
            0
        ].value
    )
    version = encoding["version"]
    input_schema = {}
    for name, schema in encoding["input_schema"].items():
        if schema["author"] != "ndonnx":
            raise ValueError(f"Unknown author {schema}")
        input_schema[name] = _get_dtype(schema["type_name"], version)

    output_schema = {}
    for name, schema in encoding["output_schema"].items():
        if schema["author"] != "ndonnx":
            raise ValueError(f"Unknown author {schema}")
        output_schema[name] = _get_dtype(schema["type_name"], version)

    return input_schema, output_schema


def get_numpy_array_api_namespace():
    if np.__version__ < "2":
        import numpy.array_api as npx

        return npx
    else:
        return np



# Copyright (c) QuantCo 2023-2024
# SPDX-License-Identifier: BSD-3-Clause

# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#
# import os
# import sys
# sys.path.insert(0, os.path.abspath('.'))


# -- Project information -----------------------------------------------------

import datetime
import importlib
import inspect
import os
import subprocess
import sys
from subprocess import CalledProcessError
from typing import cast

_mod = importlib.import_module("ndonnx")


project = "ndonnx"
copyright = f"{datetime.date.today().year}, QuantCo, Inc"
author = "QuantCo, Inc."

extensions = [
    "numpydoc",
    "sphinxcontrib.apidoc",
    "sphinx_toolbox.collapse",
    "sphinx.ext.viewcode",
    "sphinx.ext.linkcode",
    "sphinx.ext.autosummary",
    "sphinx.ext.autodoc",
    "nbsphinx",
]

apidoc_module_dir = "../ndonnx"
apidoc_output_dir = "api"
apidoc_separate_modules = True
apidoc_extra_args = ["--implicit-namespaces", "--follow-links"]

templates_path = ["_templates"]
exclude_patterns = [
    "_build",
    "Thumbs.db",
    ".DS_Store",
]
html_theme = "sphinx_rtd_theme"
html_static_path = ["_static"]

autosummary_generate = True
autodoc_type_aliases = {"npt.ArrayLike": "numpy.typing.ArrayLike"}
numpydoc_show_class_members = False
numpydoc_class_members_toctree = False


class ObjectInfo:
    def __init__(self, module_name, full_name):
        self.module_name = module_name
        self.full_name = full_name

    def linkcode_resolve(self):
        _submodule = sys.modules.get(self.module_name)
        if _submodule is None:
            return None

        _object = _submodule
        for _part in self.full_name.split("."):
            try:
                _object = getattr(_object, _part)
            except AttributeError:
                return None

        try:
            fn = inspect.getsourcefile(inspect.unwrap(_object))  # type: ignore
        except TypeError:
            fn = None
        if not fn:
            return None

        try:
            source, line_number = inspect.getsourcelines(_object)
        except OSError:
            line_number = None  # type: ignore

        if line_number:
            linespec = f"#L{line_number}-L{line_number + len(source) - 1}"
        else:
            linespec = ""

        fn = os.path.relpath(fn, start=os.path.dirname(cast(str, _mod.__file__)))

        try:
            # See https://stackoverflow.com/a/21901260
            commit = (
                subprocess.check_output(["git", "rev-parse", "HEAD"])
                .decode("ascii")
                .strip()
            )
        except CalledProcessError:
            # If subprocess returns non-zero exit status
            commit = "main"

        return (
            "https://github.com/quantco/ndonnx"
            f"/blob/{commit}/{_mod.__name__.replace('.', '/')}/{fn}{linespec}"
        )


# Copied and adapted from
# https://github.com/pandas-dev/pandas/blob/4a14d064187367cacab3ff4652a12a0e45d0711b/doc/source/conf.py#L613-L659
# Required configuration function to use sphinx.ext.linkcode
def linkcode_resolve(domain, info):
    if domain != "py":
        return None
    return ObjectInfo(info["module"], info["fullname"]).linkcode_resolve()



# Copyright (c) QuantCo 2023-2024
# SPDX-License-Identifier: BSD-3-Clause

from __future__ import annotations

import re

import numpy as np
import pytest
from typing_extensions import Self

import ndonnx as ndx
from ndonnx import (
    Array,
    CastError,
    CoreType,
)
from ndonnx._experimental import CastMixin, Schema, StructType, UniformShapeOperations


class Unsigned96Impl(UniformShapeOperations):
    def equal(self, x, y):
        return custom_equal(x, y)

    def arange(self, start, stop=None, step=None, dtype=None, device=None) -> Array:
        # Pretty naive implementation
        if stop is None or step != 1 or stop > ndx.iinfo(ndx.uint64).max:
            raise ValueError("Unsupported arange arguments")

        return Array._from_fields(
            Unsigned96(),
            upper=ndx.asarray(0, dtype=ndx.uint64),
            lower=ndx.arange(start, stop, step, dtype=ndx.uint32),
        )

    def eye(self, n_rows, n_cols=None, k=0, dtype=None, device=None) -> Array:
        return Array._from_fields(
            Unsigned96(),
            upper=ndx.asarray(0, dtype=ndx.uint64),
            lower=ndx.eye(n_rows, n_cols, k, dtype=ndx.int64, device=device).astype(
                ndx.uint32
            ),
        )

    def zeros(self, shape, dtype: CoreType | StructType | None = None, device=None):
        return ndx.full(shape, 0, dtype=dtype, device=device)

    def ones(self, shape, dtype: CoreType | StructType | None = None, device=None):
        return ndx.full(shape, 1, dtype=dtype, device=device)

    def empty(self, shape, dtype=None, device=None) -> Array:
        return ndx.zeros(shape, dtype=Unsigned96(), device=device)

    def full(self, shape, fill_value, dtype=None, device=None) -> Array:
        if not isinstance(fill_value, int):
            return NotImplemented
        if fill_value >= np.iinfo(np.uint32).max:
            raise ValueError("Unsupported fill value")
        return Array._from_fields(
            Unsigned96(),
            upper=ndx.zeros(shape, dtype=ndx.uint64, device=device),
            lower=ndx.full(shape, fill_value, dtype=ndx.uint32, device=device),
        )

    def add(self, x, y) -> Array:
        if x.dtype == Unsigned96() and y.dtype == Unsigned96():
            max_u32 = ndx.iinfo(ndx.uint32).max
            overflow_amount = (
                x.lower.astype(ndx.uint64) + y.lower.astype(ndx.uint64)
            ) >> 32
            will_overflow = (overflow_amount > 0).astype(ndx.uint32)
            # No where implemented on unsigned types.
            lower = ((x.lower + y.lower) * (1 - will_overflow)) + (
                max_u32 * will_overflow
            )
            upper = x.upper + y.upper + overflow_amount
            return Array._from_fields(
                Unsigned96(),
                upper=upper,
                lower=lower,
            )
        elif x.dtype == ndx.uint32 and y.dtype == Unsigned96():
            return y + x
        elif x.dtype == Unsigned96() and y.dtype == ndx.uint32:
            return x + y.astype(Unsigned96())
        return NotImplemented


class Unsigned96(StructType, CastMixin):
    def _fields(self) -> dict[str, StructType | CoreType]:
        return {
            "upper": ndx.uint64,
            "lower": ndx.uint32,
        }

    def _parse_input(self, x: np.ndarray) -> dict:
        upper = self._fields()["upper"]._parse_input(
            np.array(x >> 32).astype(np.uint64)
        )
        lower = self._fields()["lower"]._parse_input(
            (x & np.array([0xFFFFFFFF])).astype(np.uint32)
        )
        return {
            "upper": upper,
            "lower": lower,
        }

    def _assemble_output(self, fields: dict[str, np.ndarray]) -> np.ndarray:
        return (fields["upper"].astype(object) << 32) | fields["lower"].astype(object)

    def copy(self) -> Self:
        return self

    def _schema(self) -> Schema:
        return Schema(type_name="u96", author="value from data!")

    def _cast_to(self, array: Array, dtype: CoreType | StructType) -> Array:
        raise CastError(f"Cannot cast {self} to {dtype}")

    def _cast_from(self, array: Array) -> Array:
        if array.dtype == ndx.uint32:
            array = array.astype(ndx.uint32)
        else:
            raise CastError(f"Cannot cast {array.dtype} to {self}")

        return Array._from_fields(
            Unsigned96(), upper=ndx.asarray(0, dtype=ndx.uint64), lower=array
        )

    _ops = Unsigned96Impl()


def custom_equal(x: Array, y: Array) -> Array:
    if x.dtype != Unsigned96() or y.dtype != Unsigned96():
        raise ValueError("Can only compare Unsigned96 arrays")
    # ort doesn't implement equality on unsigned 64s, so I roll my own!
    return ((x.lower ^ y.lower) | (x.upper ^ y.upper)) == 0


@pytest.fixture
def u96():
    return Unsigned96()


def test_custom_dtype_array_creation(u96):
    ndx.asarray(np.array(0, dtype=object), dtype=u96)


def test_unsigned96_casting(u96):
    with pytest.raises(CastError):
        Array._from_fields(
            u96,
            upper=ndx.asarray(0, dtype=ndx.uint64),
            lower=ndx.asarray(0, dtype=ndx.uint32),
        ).astype(ndx.uint64)

    expected = ndx.asarray(np.array(0, dtype=object), dtype=u96)
    actual = ndx.asarray(0, dtype=ndx.uint32).astype(u96)
    custom_equal_result = custom_equal(expected, actual).to_numpy()
    assert custom_equal_result is not None and custom_equal_result.item()

    expected = ndx.asarray(np.array([1 << 40], dtype=object), dtype=u96)
    actual = ndx.asarray(np.array([1 << 40]), dtype=ndx.uint32).astype(u96)

    # Should overflow
    custom_equal_result = custom_equal(expected, actual).to_numpy()
    assert custom_equal_result is not None and not custom_equal_result.item()
    actual_value = actual.to_numpy()
    assert actual_value is not None and actual_value.item() == 0
    expected_value = expected.to_numpy()
    assert expected_value is not None and expected_value.item() == 1 << 40


def test_custom_dtype_function_dispatch(u96):
    x = ndx.asarray(np.array([22314, 21 << 12, 12], dtype=object), dtype=u96)
    y = ndx.asarray(np.array([22314, 21 << 12, 12], dtype=object), dtype=u96)
    z = ndx.asarray(np.array([223, 21 << 12, 13], dtype=object), dtype=u96)

    np.testing.assert_equal(ndx.equal(x, y).to_numpy(), [True, True, True])
    np.testing.assert_equal(ndx.equal(x, z).to_numpy(), [False, True, False])

    a = ndx.asarray(np.array([10, 0, 100], dtype=object), dtype=u96)
    np.testing.assert_equal((x + a).to_numpy(), (a + x).to_numpy())

    b = ndx.asarray(np.array([10, 2, 100], dtype=np.uint32))
    np.testing.assert_equal((x + b).to_numpy(), (b + x).to_numpy())


def test_error_message_unimplemented_dtype_dispatch(u96):
    x = ndx.asarray(np.array([22314, 21 << 12, 12], dtype=object), dtype=u96)
    y = ndx.asarray(np.array([223, 21 << 12, 13], dtype=object), dtype=u96)

    with pytest.raises(
        ndx.UnsupportedOperationError,
        match=re.escape("Unsupported operand type(s) for less_equal: 'Unsigned96'"),
    ):
        ndx.less_equal(x, y)
    with pytest.raises(
        ndx.UnsupportedOperationError,
        match=re.escape("Unsupported operand type for bitwise_invert: 'Unsigned96'"),
    ):
        ~x


def test_isinstance_i32():
    assert isinstance(ndx.int32, ndx.CoreType)
    assert isinstance(ndx.int32, ndx.Numerical)
    assert isinstance(ndx.int32, ndx.Integral)
    assert not isinstance(ndx.int32, ndx.Floating)


def test_isinstance_f32():
    assert isinstance(ndx.float32, ndx.CoreType)
    assert isinstance(ndx.float32, ndx.Numerical)
    assert isinstance(ndx.float32, ndx.Floating)
    assert not isinstance(ndx.float32, ndx.NullableIntegral)


def test_equality():
    assert ndx.int32 == ndx.int32
    assert ndx.int32 != ndx.int64
    assert ndx.int32 != ndx.nint32


def test_custom_dtype_layout_transformations(u96):
    x = ndx.asarray(np.array([22314, 21 << 12, 12, 1242134], dtype=object), dtype=u96)
    x_value = x.to_numpy()
    expected = x_value.reshape(2, 2)
    actual = ndx.reshape(x, (2, 2)).to_numpy()
    np.testing.assert_equal(expected, actual)

    expected = ndx.roll(x, 1).to_numpy()
    actual = np.roll(x.to_numpy(), 1)
    np.testing.assert_equal(expected, actual)


def test_custom_dtype_incapable_of_complex_dispatch():
    # We might consider enabling this once we have a good
    # plan for dispatching in an intuitive way where you
    # don't shoot yourself in the foot.

    x = ndx.asarray(np.array([22314, 21 << 12, 12], dtype=object), dtype=Unsigned96())
    y = ndx.asarray(np.array([223, 21 << 12, 13], dtype=object), dtype=Unsigned96())

    with pytest.raises(ndx.UnsupportedOperationError):
        ndx.concat([x, y], axis=0)


def test_custom_dtype_capable_creation_functions():
    x = ndx.full((2, 2), 0, dtype=Unsigned96())
    assert x.shape == (2, 2)
    np.testing.assert_equal(x.to_numpy(), np.array([[0, 0], [0, 0]], dtype=object))

    x = ndx.arange(0, 4, 1, dtype=Unsigned96())
    np.testing.assert_equal(x.to_numpy(), np.array([0, 1, 2, 3], dtype=object))

    x = ndx.zeros((2, 3, 2), dtype=Unsigned96())
    np.testing.assert_equal(x.to_numpy(), np.zeros((2, 3, 2), dtype=object))

    x = ndx.ones((2, 3, 2), dtype=Unsigned96())
    np.testing.assert_equal(x.to_numpy(), np.ones((2, 3, 2), dtype=object))

    x = ndx.empty((2, 3, 2), dtype=Unsigned96())
    assert x.shape == (2, 3, 2)

    x = ndx.eye(3, dtype=Unsigned96())
    np.testing.assert_equal(x.to_numpy(), np.eye(3, dtype=object))

    arr = np.array([22314, 21 << 12, 12], dtype=object)
    x = ndx.asarray(arr, dtype=Unsigned96())
    np.testing.assert_equal(
        ndx.zeros_like(x).to_numpy(), np.zeros_like(arr, dtype=object)
    )

    np.testing.assert_equal(
        ndx.ones_like(x, dtype=ndx.int32).to_numpy(), np.ones_like(arr, dtype=np.int32)
    )



# Copyright (c) QuantCo 2023-2024
# SPDX-License-Identifier: BSD-3-Clause

from __future__ import annotations

from abc import abstractmethod
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from ndonnx._array import Array
    from ndonnx._data_types import CoreType, StructType


class CastError(TypeError):
    """Error raised when a casting operation fails."""


class CastMixin:
    """Abstract interface for casting operations.

    When implemented for a data type, arrays of this data type can be casted to and from
    this data type.
    """

    @abstractmethod
    def _cast_from(self, array: Array) -> Array:
        """Casts the input array to this data type.

        Raise a `CastFailedError` if the cast is not possible.
        """
        ...

    @abstractmethod
    def _cast_to(self, array: Array, dtype: CoreType | StructType) -> Array:
        """Cast the input array from this data type to the specified data type `dtype`.

        Raise a `CastFailedError` if the cast is not possible.
        """
        ...



# Copyright (c) QuantCo 2023-2024
# SPDX-License-Identifier: BSD-3-Clause

import sys

import numpy as np
import pytest

import ndonnx as ndx
import ndonnx.additional as nda

from .utils import run


@pytest.mark.parametrize(
    "side",
    [
        "left",
        "right",
    ],
)
def test_searchsorted(side):
    a_val = [0, 1, 2, 5, 5, 6, 10, 15]
    b_val = [0, 1, 2, 3, 4, 5, 6, 7, 8, 10, 10, 15, 20, 20]
    c_val = np.searchsorted(a_val, b_val, side)

    a = ndx.asarray(a_val, dtype=ndx.int64)
    b = ndx.asarray(b_val, dtype=ndx.int64)
    c = ndx.searchsorted(a, b, side=side)
    np.testing.assert_equal(c_val, c.to_numpy())


@pytest.mark.skip(reason="TODO: onnxruntime")
@pytest.mark.parametrize(
    "side",
    [
        "left",
        "right",
    ],
)
def test_searchsorted_nans(side):
    a_val = np.array([0, 1, 2, 5, 5, 6, 10, 15, np.nan])
    b_val = np.array([0, 1, 2, np.nan, np.nan])
    c_val = np.searchsorted(a_val, b_val, side)

    a = ndx.array(shape=(len(a_val),), dtype=ndx.float64)
    b = ndx.array(shape=(len(b_val),), dtype=ndx.float64)
    c = ndx.searchsorted(a, b, side=side)

    model = ndx.build({"a": a, "b": b}, {"c": c})

    np.testing.assert_equal(c_val, run(model, dict(a=a_val, b=b_val))["c"])


def test_searchsorted_raises():
    with pytest.raises(TypeError):
        a = ndx.array(shape=(), dtype=ndx.int64)
        b = ndx.array(shape=(), dtype=ndx.float64)

        ndx.searchsorted(a, b)

    with pytest.raises(ValueError):
        a = ndx.array(shape=(3,), dtype=ndx.int64)
        b = ndx.array(shape=(3,), dtype=ndx.int64)

        ndx.searchsorted(a, b, side="middle")  # type: ignore[arg-type]


@pytest.mark.skipif(
    sys.platform.startswith("win"),
    reason="ORT 1.18 not registering LabelEncoder(4) only on Windows.",
)
def test_static_map():
    a = ndx.array(shape=(3,), dtype=ndx.int64)
    b = nda.static_map(a, {1: 2, 2: 3})

    model = ndx.build({"a": a}, {"b": b})
    np.testing.assert_equal([0, 2, 3], run(model, {"a": np.array([0, 1, 2])})["b"])

    # nans are mapped by static_map
    a = ndx.array(shape=("N",), dtype=ndx.float64)
    b = nda.static_map(a, {1.0: 2, np.nan: 3, 3.0: 42}, default=-1)

    model = ndx.build({"a": a}, {"b": b})
    np.testing.assert_equal(
        [-1, -1, 42, 3],
        run(model, {"a": np.array([0.0, 2.0, 3.0, np.nan])})["b"],
    )


@pytest.mark.skipif(
    sys.platform.startswith("win"),
    reason="ORT 1.18 not registering LabelEncoder(4) only on Windows.",
)
def test_isin():
    a = ndx.array(shape=("N",), dtype=ndx.utf8)
    b = nda.isin(a, ["foo", "bar", "baz"])

    model = ndx.build({"a": a}, {"b": b})
    np.testing.assert_equal(
        [False, True, True, False],
        run(model, dict(a=np.array(["hello", "foo", "baz", "!"])))["b"],
    )

    a = ndx.array(shape=("N",), dtype=ndx.utf8)
    b = nda.isin(a, ["🔴", "🟡", "🟢"])

    model = ndx.build({"a": a}, {"b": b})
    np.testing.assert_equal(
        [False, True, False],
        run(model, dict(a=np.array(["🚀", "🔴", "hi🟡"])))["b"],
    )

    a = ndx.asarray(["hello", "world"])
    np.testing.assert_equal([True, False], nda.isin(a, ["hello"]).to_numpy())



# Copyright (c) QuantCo 2023-2024
# SPDX-License-Identifier: BSD-3-Clause


def test_import():
    import ndonnx  # noqa


def test_onnxruntime():
    import onnxruntime

    # Test that the onnxruntime has InfereceSession
    assert hasattr(onnxruntime, "InferenceSession")



# Copyright (c) QuantCo 2023-2024
# SPDX-License-Identifier: BSD-3-Clause

from __future__ import annotations

import re

import numpy as np
import pytest
import spox.opset.ai.onnx.v19 as op

import ndonnx as ndx
import ndonnx.additional as nda
from ndonnx import _data_types as dtypes
from ndonnx._utility import promote

from .utils import get_numpy_array_api_namespace, run


def numpy_to_graph_input(arr, eager=False):
    dtype: dtypes.CoreType | dtypes.StructType
    if isinstance(arr, np.ma.MaskedArray):
        dtype = dtypes.into_nullable(dtypes.from_numpy_dtype(arr.dtype))
    else:
        dtype = dtypes.from_numpy_dtype(arr.dtype)
    return (
        ndx.array(
            shape=arr.shape,
            dtype=dtype,
        )
        if not eager
        else ndx.asarray(
            arr,
            dtype=dtype,
        )
    )


@pytest.fixture
def _a():
    return np.array([1, 2, 3])


@pytest.fixture
def _b():
    return np.array([2, 3, 3])


@pytest.fixture
def _c():
    return np.array([[1, 2, 3], [4, 5, 6]])


@pytest.mark.parametrize(
    "dtype, input",
    [
        (ndx.bool, np.array([True, False, True])),
        (ndx.int8, np.array([1, 5, -12], np.int8)),
        (ndx.nbool, np.ma.masked_array([True, True, False], mask=[1, 0, 0])),
        (ndx.nint8, np.ma.masked_array([1, 5, -12], mask=[0, 1, 1], dtype=np.int8)),
    ],
)
@pytest.mark.parametrize("shape", [(3,), ("N",)])
def test_make_graph_input(dtype, input, shape):
    a = ndx.array(shape=shape, dtype=dtype)
    model = ndx.build({"a": a}, {"b": a})
    actual = run(model, {"a": input})
    np.testing.assert_equal(input, actual["b"])


def test_null_promotion():
    a = ndx.array(shape=("N",), dtype=ndx.nfloat64)
    b = ndx.array(shape=("N",), dtype=ndx.float64)
    model = ndx.build({"a": a, "b": b}, {"c": a + b})
    inputs = {
        "a": np.ma.masked_array([1.0, 2.0, 3.0, 4.0], mask=[0, 0, 1, 0]),
        "b": np.array([4.0, 5.0, 6.0, 7.0]),
    }
    actual = run(model, inputs)
    np.testing.assert_equal(np.add(inputs["a"], inputs["b"]), actual["c"])


def test_asarray():
    a = ndx.asarray([1, 2, 3], dtype=ndx.int64)
    assert a.dtype == ndx.int64
    np.testing.assert_array_equal(
        np.array([1, 2, 3], np.int64), a.to_numpy(), strict=True
    )


def test_asarray_masked():
    np_arr = np.ma.masked_array([1, 2, 3], mask=[0, 0, 1])
    ndx_arr = ndx.asarray(np_arr)
    assert isinstance(ndx_arr, ndx.Array)
    assert isinstance(ndx_arr.to_numpy(), np.ma.MaskedArray)
    np.testing.assert_array_equal(np_arr, ndx_arr.to_numpy())


def test_basic_eager_add(_a, _b):
    a = ndx.asarray(_a)
    b = ndx.asarray(_b)
    x = ndx.asarray([1]) + 2
    c = a + b + x
    np.testing.assert_array_equal(c.to_numpy(), _a + _b + np.array([1]) + 2)


def test_string_concatenation():
    a = ndx.asarray(["a", "b", "c"])
    b = ndx.asarray(["d", "e", "f"])
    np.testing.assert_array_equal((a + b).to_numpy(), np.array(["ad", "be", "cf"]))
    np.testing.assert_array_equal((a + "d").to_numpy(), np.array(["ad", "bd", "cd"]))


def test_combining_lazy_eager():
    a = ndx.array(shape=(3,), dtype=ndx.int64)
    b = ndx.asarray(np.array([1, 2, 3], dtype=np.int64))
    c = a + b
    assert not c.to_numpy() is not None


def test_basic_indexing():
    a = ndx.asarray([1, 2, 3])
    b = a[0]
    a[0] = 2
    assert (a[0] != b).to_numpy().item()
    assert (a[0] == 2).to_numpy().item()


def test_lazy_array(_a, _b):
    a = numpy_to_graph_input(_a)
    b = ndx.asarray(_b)
    c = a * b

    model = ndx.build({"a": a}, {"c": c})
    actual = run(model, {"a": _a})["c"]
    expected = _a * _b
    np.testing.assert_equal(actual, expected)


def test_indexing(_a):
    a = numpy_to_graph_input(_a)
    b = a[0]

    model = ndx.build({"a": a}, {"b": b})
    actual = run(model, {"a": _a})["b"]
    expected = _a[0]
    np.testing.assert_array_equal(expected, actual)


@pytest.mark.parametrize(
    "condition, x, y",
    [
        (
            np.array([True, False, False], dtype=np.bool_),
            np.array([1, 2, 3]),
            np.array([3.12, 3.24, -124.0]),
        ),
        (
            np.array([True, False, True], dtype=np.bool_),
            np.ma.masked_array([1, 2, 3], mask=[0, 0, 1]),
            np.array([3.12, 3.24, -124.0]),
        ),
    ],
)
def test_where(condition, x, y):
    expected = np.where(condition, x, y)
    condition_onnx = numpy_to_graph_input(condition)
    x_onnx = numpy_to_graph_input(x)
    y_onnx = numpy_to_graph_input(y)
    out = ndx.where(condition_onnx, x_onnx, y_onnx)
    model = ndx.build(
        {"condition": condition_onnx, "x": x_onnx, "y": y_onnx}, {"out": out}
    )
    actual = run(model, {"condition": condition, "x": x, "y": y})["out"]
    np.testing.assert_array_equal(expected, actual)


def test_indexing_on_scalar():
    res = ndx.asarray(1)
    res = res[()]
    res[()] = 2
    assert res == 2


def test_indexing_on_scalar_mask():
    res = ndx.asarray([])
    res = res[False]
    np.testing.assert_equal(nda.shape(res).to_numpy(), (0, 0))


def test_indexing_with_mask(_a):
    _mask = np.array([True, False, True])

    a = numpy_to_graph_input(_a)
    mask = numpy_to_graph_input(_mask)
    c = ndx.reshape(a[mask], [-1])

    expected_c = _a[_mask]

    model = ndx.build({"a": a, "mask_": mask}, {"c": c})
    actual = run(model, {"a": _a, "mask_": _mask})["c"]
    np.testing.assert_array_equal(expected_c, actual)


def test_indexing_with_mask_raises(_a):
    a = numpy_to_graph_input(_a)
    mask = ndx.array(shape=(3, 1, 1), dtype=ndx.bool)

    with pytest.raises(IndexError):
        a[mask]


def test_indexing_with_array(_a):
    _index = np.array([0, 2])

    a = numpy_to_graph_input(_a)
    index = numpy_to_graph_input(_index)
    c = ndx.reshape(a[index], [-1])

    expected_c = _a[_index]

    model = ndx.build({"a": a, "index_": index}, {"c": c})
    actual = run(model, {"a": _a, "index_": _index})["c"]
    np.testing.assert_array_equal(expected_c, actual)


def test_indexing_with_tuple_of_array(_a):
    a = numpy_to_graph_input(_a)
    index = numpy_to_graph_input(np.array([0, 2]))

    with pytest.raises(
        TypeError,
        match=re.escape(f"Index {index} for type {type(index)} not supported"),
    ):
        a[(index,)]


def test_slicing(_a):
    a = numpy_to_graph_input(_a)
    b = a[1:2]

    model = ndx.build({"a": a}, {"b": b})
    actual = run(model, {"a": _a})["b"]
    np.testing.assert_equal(actual, _a[1:2])


def test_indexing_list(_a):
    a = numpy_to_graph_input(_a)

    with pytest.raises(TypeError, match="not supported"):
        a[[0, 2]]


def test_indexing_slice_ellipsis(_c):
    a = numpy_to_graph_input(_c)
    b = a[..., 1]

    model = ndx.build({"a": a}, {"b": b})
    actual = run(model, {"a": _c})["b"]
    expected = _c[..., 1]
    np.testing.assert_array_equal(actual, expected)


def test_indexing_none(_c):
    _index = (None, 1, None, 1)

    a = numpy_to_graph_input(_c)
    b = a[_index]

    model = ndx.build({"a": a}, {"b": b})
    actual = run(model, {"a": _c})["b"]
    expected = _c[_index]
    np.testing.assert_array_equal(actual, expected)


def test_illegal_indexing(_a):
    a = numpy_to_graph_input(_a)

    with pytest.raises(TypeError):
        a["invalid", ...]


def test_indexing_with_invalid_rank(_a):
    a = numpy_to_graph_input(_a)

    with pytest.raises(IndexError):
        a[()]

    b = numpy_to_graph_input(np.array([[1, 2], [3, 4]]))
    with pytest.raises(IndexError):
        b[0, 0, 0]
    with pytest.raises(IndexError):
        b[0,]

    b[0, ...]


def test_indexing_set_with_array(_a):
    _index = np.array([0, 2])

    a = numpy_to_graph_input(_a)
    index = numpy_to_graph_input(_index)

    c = a.copy()
    c[index] = 0

    expected_c = _a
    expected_c[_index] = 0

    model = ndx.build({"a": a, "index_": index}, {"c": c})
    actual = run(model, {"a": _a, "index_": _index})["c"]
    np.testing.assert_equal(actual, expected_c)


def test_indexing_set_scalar():
    _a = np.array(2)

    a = numpy_to_graph_input(_a)

    c = a.copy()
    c = ndx.asarray(0)

    expected_c = 0

    model = ndx.build({"a": a}, {"c": c})
    np.testing.assert_array_equal(expected_c, run(model, {"a": _a})["c"], strict=True)


# Check if repr does not raise
def test_repr():
    a = ndx.array(shape=(3,), dtype=ndx.int64)
    repr(a)

    b = ndx.asarray([1, 2, 3])
    repr(b)

    c = ndx.array(shape=(3,), dtype=ndx.nint8)
    repr(c)

    d = ndx.asarray(np.array([1, 2, 3], np.int64))
    repr(d)


@pytest.mark.parametrize(
    "rop", ["__radd__", "__rsub__", "__rmul__", "__rtruediv__", "__rpow__", "__rmod__"]
)
def test_rops(rop):
    a = ndx.asarray([1.0, 2.0, 3.0])
    b = ndx.asarray([1.0, 2.0, 3.0])

    res = getattr(a, rop)(b)
    np.testing.assert_equal(res.to_numpy(), getattr(a.to_numpy(), rop)(b.to_numpy()))


def test_rsub():
    a = ndx.asarray([1, 2, 3])

    res = 1 - a
    a_value = a.to_numpy()
    assert a_value is not None
    np.testing.assert_equal(res.to_numpy(), 1 - a_value)


def test_matrix_transpose():
    a = ndx.array(shape=(3, 2, 3), dtype=ndx.int64)
    b = ndx.matrix_transpose(a)

    model = ndx.build({"a": a}, {"b": b})
    npx = get_numpy_array_api_namespace()
    np.testing.assert_equal(
        npx.matrix_transpose(npx.reshape(npx.arange(3 * 2 * 3), (3, 2, 3))),
        run(model, {"a": np.arange(3 * 2 * 3, dtype=np.int64).reshape(3, 2, 3)})["b"],
    )


def test_matrix_transpose_attribute():
    a = ndx.array(shape=(3, 2, 3), dtype=ndx.int64)
    b = a.mT

    model = ndx.build({"a": a}, {"b": b})
    npx = get_numpy_array_api_namespace()
    expected = npx.reshape(npx.arange(3 * 2 * 3), (3, 2, 3)).mT

    np.testing.assert_equal(
        expected,
        run(model, {"a": np.arange(3 * 2 * 3, dtype=np.int64).reshape(3, 2, 3)})["b"],
    )


def test_transpose_attribute():
    a = ndx.array(shape=(3, 2), dtype=ndx.int64)
    b = a.T

    model = ndx.build({"a": a}, {"b": b})
    np.testing.assert_equal(
        np.reshape(np.arange(3 * 2), (3, 2)).T,
        run(model, {"a": np.arange(3 * 2, dtype=np.int64).reshape(3, 2)})["b"],
    )


def test_array_spox_interoperability():
    a = ndx.array(shape=(3, 2), dtype=ndx.nint64)
    add_var = op.add(a.values.data.var, op.const(5, dtype=np.int64))  # type: ignore
    b = ndx.from_spox_var(var=add_var)
    model = ndx.build({"a": a}, {"b": b})
    expected = np.reshape(np.arange(3 * 2), (3, 2)) + 5
    input = np.ma.masked_array(
        np.arange(3 * 2, dtype=np.int64).reshape(3, 2), mask=np.ones((3, 2), dtype=bool)
    )
    actual = run(model, {"a": input})["b"]
    np.testing.assert_equal(expected, actual)


def test_creation_arange():
    a = ndx.arange(0, stop=10)
    np.testing.assert_equal(a.to_numpy(), np.arange(stop=10))

    b = ndx.arange(1, 10)
    np.testing.assert_equal(b.to_numpy(), np.arange(1, 10))

    c = ndx.arange(1, 10, 2)
    np.testing.assert_equal(c.to_numpy(), np.arange(1, 10, 2))

    d = ndx.arange(0.0, None, step=-1)
    np.testing.assert_array_equal(
        np.arange(0.0, None, step=-1), d.to_numpy(), strict=True
    )


def test_creation_full():
    a = ndx.full((2, 3), 5)
    np.testing.assert_equal(np.full((2, 3), 5), a.to_numpy())

    b = ndx.full((2, 3), 5, dtype=ndx.float32)
    np.testing.assert_equal(np.full((2, 3), 5, dtype=np.float32), b.to_numpy())
    c = ndx.full((2, 3), "a", dtype=ndx.nutf8)
    np.testing.assert_equal(np.full((2, 3), "a"), c.to_numpy())

    d = ndx.full(2, 5, dtype=ndx.int8)
    np.testing.assert_equal(np.full(2, 5, dtype=np.int8), d.to_numpy())

    # Check lazy creation
    e = ndx.array(shape=tuple(), dtype=ndx.int64)
    f = ndx.full(e, 10)
    model_proto = ndx.build({"e": e}, {"f": f})
    actual = run(model_proto, {"e": np.array(5, dtype=np.int64)})["f"]
    np.testing.assert_equal(np.array([10] * 5, dtype=np.int64), actual)

    # Note we must know the output shape to export an ONNX artifact.
    g = ndx.array(shape=(2,), dtype=ndx.int64)
    h = ndx.full(g, 10)
    model_proto = ndx.build({"g": g}, {"h": h})
    np.testing.assert_equal(
        np.array([[10, 10, 10], [10, 10, 10]]),
        run(model_proto, {"g": np.array([2, 3], dtype=np.int64)})["h"],
    )


@pytest.mark.parametrize(
    "args, expected",
    [
        ((ndx.int32, ndx.int64), ndx.int64),
        (
            (ndx.asarray([1, 2, 3], ndx.int64), ndx.asarray([1, 2, 3], ndx.int32)),
            ndx.int64,
        ),
        ((ndx.int32, ndx.asarray([1, 2, 3], ndx.int64)), ndx.int64),
        ((ndx.float32, ndx.float64), ndx.float64),
        ((ndx.float64, ndx.int32), ndx.float64),
    ],
)
def test_result_type(args, expected):
    assert ndx.result_type(*args) == expected


def test_ceil():
    a = ndx.asarray(np.array([1.2, 1.3, -13.13]))
    np.testing.assert_equal(ndx.ceil(a).to_numpy(), [2.0, 2.0, -13.0])


def test_propagates_minimal_dtype():
    a = ndx.asarray([1, 2, 4], dtype=ndx.int8)
    b = a + 1
    assert b.dtype == ndx.int8
    np.testing.assert_equal(b.to_numpy(), [2, 3, 5])


@pytest.mark.parametrize(
    "x",
    [
        ndx.asarray([True, False, False]),
        ndx.asarray([False]),
        ndx.asarray([True]),
        ndx.asarray([True, True]),
        ndx.asarray([], dtype=ndx.bool),
    ],
)
def test_all(x):
    np.testing.assert_equal(ndx.all(x).to_numpy(), np.all(x.to_numpy()))


@pytest.mark.parametrize(
    "side",
    [
        "left",
        "right",
    ],
)
def test_searchsorted(side):
    a_val = [0, 1, 2, 5, 5, 6, 10, 15]
    b_val = [0, 1, 2, 3, 4, 5, 6, 7, 8, 10, 10, 15, 20, 20]
    c_val = np.searchsorted(a_val, b_val, side)

    a = ndx.asarray(a_val, dtype=ndx.int64)
    b = ndx.asarray(b_val, dtype=ndx.int64)
    c = ndx.searchsorted(a, b, side=side)
    np.testing.assert_equal(c_val, c.to_numpy())


@pytest.mark.skip(reason="TODO: onnxruntime")
@pytest.mark.parametrize(
    "side",
    [
        "left",
        "right",
    ],
)
def test_searchsorted_nans(side):
    a_val = np.array([0, 1, 2, 5, 5, 6, 10, 15, np.nan])
    b_val = np.array([0, 1, 2, np.nan, np.nan])
    c_val = np.searchsorted(a_val, b_val, side)

    a = ndx.array(shape=(len(a_val),), dtype=ndx.float64)
    b = ndx.array(shape=(len(b_val),), dtype=ndx.float64)
    c = ndx.searchsorted(a, b, side=side)

    model = ndx.build({"a": a, "b": b}, {"c": c})

    np.testing.assert_equal(c_val, run(model, dict(a=a_val, b=b_val))["c"])


def test_searchsorted_raises():
    with pytest.raises(TypeError):
        a = ndx.array(shape=(), dtype=ndx.int64)
        b = ndx.array(shape=(), dtype=ndx.float64)

        ndx.searchsorted(a, b)

    with pytest.raises(ValueError):
        a = ndx.array(shape=(3,), dtype=ndx.int64)
        b = ndx.array(shape=(3,), dtype=ndx.int64)

        ndx.searchsorted(a, b, side="middle")  # type: ignore[arg-type]


def test_truediv():
    x = ndx.asarray([1, 2, 3], dtype=ndx.int64)
    y = ndx.asarray([2, 3, 3], dtype=ndx.int64)
    z = x / y
    assert isinstance(z.dtype, ndx.Floating)
    np.testing.assert_array_equal(z.to_numpy(), np.array([0.5, 2 / 3, 1.0]))


@pytest.mark.parametrize(
    "dtype",
    [
        ndx.float32,
        ndx.nfloat32,
        ndx.int8,
        ndx.int16,
        ndx.int32,
        ndx.int64,
        ndx.nint8,
        ndx.nint16,
        ndx.nint32,
        ndx.nint64,
        ndx.uint8,
        ndx.uint16,
        ndx.uint32,
        ndx.nuint8,
        ndx.nuint16,
        ndx.nuint32,
    ],
)
def test_prod(dtype):
    x = ndx.asarray([2, 2]).astype(dtype)
    y = ndx.prod(x)

    np.testing.assert_array_equal(y.to_numpy(), np.array(4))


@pytest.mark.parametrize(
    "dtype",
    [
        ndx.float64,
        ndx.nfloat64,
        ndx.uint64,
        ndx.nuint64,
    ],
)
def test_prod_no_implementation(dtype):
    x = ndx.asarray([2, 2]).astype(dtype)
    with pytest.raises(TypeError):
        ndx.prod(x)


def test_array_creation_with_invalid_fields():
    with pytest.raises(TypeError):
        ndx.Array._from_fields(
            ndx.nutf8,
            values=ndx.array(shape=(3,), dtype=ndx.int32),
            invalid_field=ndx.array(shape=(3,), dtype=ndx.utf8),
        )

    with pytest.raises(TypeError):
        ndx.Array._from_fields(
            ndx.nutf8,
            values=ndx.array(shape=(3,), dtype=ndx.int32),
            null=ndx.array(shape=(3,), dtype=ndx.bool),
        )

    with pytest.raises(TypeError):
        ndx.Array._from_fields(
            ndx.utf8,
            values=ndx.array(shape=(3,), dtype=ndx.utf8),
        )

    with pytest.raises(TypeError):
        ndx.Array._from_fields(
            ndx.utf8,
            values=ndx.array(shape=(3,), dtype=ndx.int32)._core(),
        )


def test_promote_nullable():
    with pytest.warns(DeprecationWarning):
        assert ndx.promote_nullable(np.int64) == ndx.nint64


@pytest.mark.parametrize(
    "operation", [ndx.sin, ndx.cos, ndx.tan, ndx.sinh, ndx.mean, ndx.sum, ndx.abs]
)
@pytest.mark.parametrize("dtype", [ndx.utf8, ndx.nutf8, ndx.bool, ndx.nbool])
def test_numerical_unary_operations_fail_on_non_numeric_input(operation, dtype):
    a = ndx.array(shape=(3,), dtype=dtype)
    with pytest.raises(
        ndx.UnsupportedOperationError, match="Unsupported operand type for"
    ):
        operation(a)


def test_string_shape_operations():
    a = ndx.asarray(["a", "b", "c"])
    b = ndx.broadcast_to(a, (2, 3))
    np.testing.assert_array_equal(
        b.to_numpy(), np.array([["a", "b", "c"], ["a", "b", "c"]])
    )

    c = ndx.concat([a, a], axis=0)
    np.testing.assert_array_equal(
        c.to_numpy(), np.array(["a", "b", "c", "a", "b", "c"])
    )


@pytest.mark.parametrize(
    "dtype",
    [
        ndx.utf8,
        ndx.bool,
    ],
)
def test_zeros(dtype):
    x = ndx.zeros((2, 3), dtype=dtype)
    np.testing.assert_equal(
        x.to_numpy(), np.zeros((2, 3), dtype=dtype.to_numpy_dtype())
    )


@pytest.mark.xfail(reason="https://github.com/onnx/onnx/issues/6276")
def test_empty_concat_eager():
    a = ndx.asarray([], ndx.int64)
    b = ndx.asarray([1, 2, 3], ndx.int64)
    out = ndx.concat([a, b], axis=0)
    np.testing.assert_equal(out.to_numpy(), b.to_numpy())


def test_empty_concat_lazy_known_shape():
    a = ndx.array(shape=(0,), dtype=ndx.int64)
    b = ndx.array(shape=(3,), dtype=ndx.int64)
    out = ndx.concat([a, b])
    model = ndx.build({"a": a, "b": b}, {"out": out})

    anp = np.array([], np.int64)
    bnp = np.array([1, 2, 3], np.int64)

    out = run(model, {"a": anp, "b": bnp})["out"]

    np.testing.assert_equal(out, bnp)


def test_empty_concat_lazy_unknown_shape():
    a = ndx.array(shape=(None,), dtype=ndx.int64)
    b = ndx.array(shape=(None,), dtype=ndx.int64)
    out = ndx.concat([a, b])
    model = ndx.build({"a": a, "b": b}, {"out": out})

    anp = np.array([], np.int64)
    bnp = np.array([1, 2, 3], np.int64)

    out = run(model, {"a": anp, "b": bnp})["out"]

    np.testing.assert_equal(out, bnp)


# if the precision loss looks concerning, note https://data-apis.org/array-api/latest/API_specification/type_promotion.html#mixing-arrays-with-python-scalars
@pytest.mark.parametrize(
    "arrays, scalar",
    [
        ([ndx.array(shape=("N",), dtype=ndx.uint8)], 1),
        ([ndx.array(shape=("N",), dtype=ndx.uint8)], -1),
        ([ndx.array(shape=("N",), dtype=ndx.int8)], 1),
        ([ndx.array(shape=("N",), dtype=ndx.nint8)], 1),
        ([ndx.array(shape=("N",), dtype=ndx.nint8)], 1),
        ([ndx.array(shape=("N",), dtype=ndx.float64)], 0.123456789),
        ([ndx.array(shape=("N",), dtype=ndx.float64)], np.float64(0.123456789)),
        ([ndx.array(shape=("N",), dtype=ndx.float32)], 0.123456789),
        (
            [
                ndx.array(shape=("N",), dtype=ndx.float32),
                ndx.asarray([1.5], dtype=ndx.float64),
            ],
            0.123456789,
        ),
        ([ndx.asarray(["a", "b"], dtype=ndx.utf8)], "hello"),
    ],
)
def test_scalar_promote(arrays, scalar):
    args = arrays + [scalar]
    *updated_arrays, updated_scalar = promote(*args)
    assert all(isinstance(array, ndx.Array) for array in updated_arrays)
    assert isinstance(updated_scalar, ndx.Array)
    assert updated_scalar.shape == ()
    assert all(array.dtype == updated_scalar.dtype for array in updated_arrays)


@pytest.mark.parametrize(
    "arrays, scalar",
    [
        ([ndx.asarray(["a", "b"], dtype=ndx.utf8)], 1),
        ([ndx.asarray([1, 2], dtype=ndx.int32)], "hello"),
    ],
)
def test_promotion_failures(arrays, scalar):
    with pytest.raises(TypeError, match="Cannot promote"):
        promote(*arrays, scalar)


@pytest.mark.parametrize(
    "x, y",
    [
        (np.asarray([1, 2, 3], dtype=np.int64), 1.12),
        (1.23, np.asarray([1, 2, 3], dtype=np.int64)),
        (True, np.asarray([1, 2, 3], dtype=np.int8)),
        (np.asarray([True, False]), 1.12),
        (np.asarray([True, False]), 4),
        (np.asarray([1.23, 2.34]), True),
        (np.asarray([1.23, 2.34]), 2),
    ],
)
def test_cross_kind_promotions(x, y):
    np_result = x + y
    if isinstance(x, np.ndarray):
        x = ndx.asarray(x)
    if isinstance(y, np.ndarray):
        y = ndx.asarray(y)
    onnx_result = x + y
    np.testing.assert_equal(onnx_result.to_numpy(), np_result)



# Copyright (c) QuantCo 2023-2024
# SPDX-License-Identifier: BSD-3-Clause

from __future__ import annotations

from typing import Literal

import ndonnx as ndx
import ndonnx._data_types as dtypes


class OperationsBlock:
    """Interface for data types to implement top-level functions exported by ndonnx."""

    # elementwise.py

    def abs(self, x) -> ndx.Array:
        return NotImplemented

    def acos(self, x) -> ndx.Array:
        return NotImplemented

    def acosh(self, x) -> ndx.Array:
        return NotImplemented

    def add(self, x, y) -> ndx.Array:
        return NotImplemented

    def asin(self, x) -> ndx.Array:
        return NotImplemented

    def asinh(self, x) -> ndx.Array:
        return NotImplemented

    def atan(self, x) -> ndx.Array:
        return NotImplemented

    def atan2(self, y, x) -> ndx.Array:
        return NotImplemented

    def atanh(self, x) -> ndx.Array:
        return NotImplemented

    def bitwise_and(self, x, y) -> ndx.Array:
        return NotImplemented

    def bitwise_left_shift(self, x, y) -> ndx.Array:
        return NotImplemented

    def bitwise_invert(self, x) -> ndx.Array:
        return NotImplemented

    def bitwise_or(self, x, y) -> ndx.Array:
        return NotImplemented

    def bitwise_right_shift(self, x, y) -> ndx.Array:
        return NotImplemented

    def bitwise_xor(self, x, y) -> ndx.Array:
        return NotImplemented

    def ceil(self, x) -> ndx.Array:
        return NotImplemented

    def conj(*args):
        return NotImplemented

    def cos(self, x) -> ndx.Array:
        return NotImplemented

    def cosh(self, x) -> ndx.Array:
        return NotImplemented

    def divide(self, x, y) -> ndx.Array:
        return NotImplemented

    def equal(self, x, y) -> ndx.Array:
        return NotImplemented

    def exp(self, x) -> ndx.Array:
        return NotImplemented

    def expm1(self, x) -> ndx.Array:
        return NotImplemented

    def floor(self, x) -> ndx.Array:
        return NotImplemented

    def floor_divide(self, x, y) -> ndx.Array:
        return NotImplemented

    def greater(self, x, y) -> ndx.Array:
        return NotImplemented

    def greater_equal(self, x, y) -> ndx.Array:
        return NotImplemented

    def imag(*args):
        return NotImplemented

    def isfinite(self, x) -> ndx.Array:
        return NotImplemented

    def isinf(self, x) -> ndx.Array:
        return NotImplemented

    def isnan(self, x) -> ndx.Array:
        return NotImplemented

    def less(self, x, y) -> ndx.Array:
        return NotImplemented

    def less_equal(self, x, y) -> ndx.Array:
        return NotImplemented

    def log(self, x) -> ndx.Array:
        return NotImplemented

    def log1p(self, x) -> ndx.Array:
        return NotImplemented

    def log2(self, x) -> ndx.Array:
        return NotImplemented

    def log10(self, x) -> ndx.Array:
        return NotImplemented

    def logaddexp(self, x, y) -> ndx.Array:
        return NotImplemented

    def logical_and(self, x, y) -> ndx.Array:
        return NotImplemented

    def logical_not(self, x) -> ndx.Array:
        return NotImplemented

    def logical_or(self, x, y) -> ndx.Array:
        return NotImplemented

    def logical_xor(self, x, y) -> ndx.Array:
        return NotImplemented

    def multiply(self, x, y) -> ndx.Array:
        return NotImplemented

    def negative(self, x) -> ndx.Array:
        return NotImplemented

    def not_equal(self, x, y) -> ndx.Array:
        return NotImplemented

    def positive(self, x) -> ndx.Array:
        return NotImplemented

    def pow(self, x, y) -> ndx.Array:
        return NotImplemented

    def real(*args):
        return NotImplemented

    def remainder(self, x, y) -> ndx.Array:
        return NotImplemented

    def round(self, x) -> ndx.Array:
        return NotImplemented

    def sign(self, x) -> ndx.Array:
        return NotImplemented

    def sin(self, x) -> ndx.Array:
        return NotImplemented

    def sinh(self, x) -> ndx.Array:
        return NotImplemented

    def square(self, x) -> ndx.Array:
        return NotImplemented

    def sqrt(self, x) -> ndx.Array:
        return NotImplemented

    def subtract(self, x, y) -> ndx.Array:
        return NotImplemented

    def tan(self, x) -> ndx.Array:
        return NotImplemented

    def tanh(self, x) -> ndx.Array:
        return NotImplemented

    def trunc(self, x) -> ndx.Array:
        return NotImplemented

    # linalg.py

    def matmul(self, x, y) -> ndx.Array:
        return NotImplemented

    def matrix_transpose(self, x) -> ndx.Array:
        return NotImplemented

    # searching.py

    def argmax(self, x, axis=None, keepdims=False) -> ndx.Array:
        return NotImplemented

    def argmin(self, x, axis=None, keepdims=False) -> ndx.Array:
        return NotImplemented

    def nonzero(self, x) -> tuple[ndx.Array, ...]:
        return NotImplemented

    def searchsorted(
        self,
        x1,
        x2,
        *,
        side: Literal["left", "right"] = "left",
        sorter: ndx.Array | None = None,
    ) -> ndx.Array:
        return NotImplemented

    def where(self, condition, x, y) -> ndx.Array:
        return NotImplemented

    # set.py
    def unique_all(self, x):
        return NotImplemented

    def unique_counts(self, x):
        return NotImplemented

    def unique_inverse(self, x):
        return NotImplemented

    def unique_values(self, x):
        return NotImplemented

    # sorting.py

    def argsort(self, x, *, axis=-1, descending=False, stable=True) -> ndx.Array:
        return NotImplemented

    def sort(self, x, *, axis=-1, descending=False, stable=True) -> ndx.Array:
        return NotImplemented

    # statistical.py

    def cumulative_sum(
        self,
        x,
        *,
        axis: int | None = None,
        dtype: ndx.CoreType | dtypes.StructType | None = None,
        include_initial: bool = False,
    ):
        return NotImplemented

    def max(self, x, *, axis=None, keepdims: bool = False) -> ndx.Array:
        return NotImplemented

    def mean(self, x, *, axis=None, keepdims: bool = False) -> ndx.Array:
        return NotImplemented

    def min(self, x, *, axis=None, keepdims: bool = False) -> ndx.Array:
        return NotImplemented

    def prod(
        self,
        x,
        *,
        axis=None,
        dtype: dtypes.CoreType | dtypes.StructType | None = None,
        keepdims: bool = False,
    ) -> ndx.Array:
        return NotImplemented

    def clip(self, x, *, min=None, max=None) -> ndx.Array:
        return NotImplemented

    def std(
        self,
        x,
        axis=None,
        keepdims: bool = False,
        correction=0.0,
        dtype: dtypes.StructType | dtypes.CoreType | None = None,
    ) -> ndx.Array:
        return NotImplemented

    def sum(
        self,
        x,
        *,
        axis=None,
        dtype: dtypes.CoreType | dtypes.StructType | None = None,
        keepdims: bool = False,
    ) -> ndx.Array:
        return NotImplemented

    def var(
        self,
        x,
        *,
        axis=None,
        keepdims: bool = False,
        correction=0.0,
        dtype: dtypes.CoreType | dtypes.StructType | None = None,
    ) -> ndx.Array:
        return NotImplemented

    # utility.py

    def all(self, x, *, axis=None, keepdims: bool = False) -> ndx.Array:
        return NotImplemented

    def any(self, x, *, axis=None, keepdims: bool = False) -> ndx.Array:
        return NotImplemented

    # indexing.py
    def take(self, x, indices, axis=None) -> ndx.Array:
        return NotImplemented

    def broadcast_to(self, x, shape) -> ndx.Array:
        return NotImplemented

    def expand_dims(self, x, axis) -> ndx.Array:
        return NotImplemented

    def flip(self, x, axis) -> ndx.Array:
        return NotImplemented

    def permute_dims(self, x, axes) -> ndx.Array:
        return NotImplemented

    def reshape(self, x, shape, *, copy=None) -> ndx.Array:
        return NotImplemented

    def roll(self, x, shift, axis) -> ndx.Array:
        return NotImplemented

    def squeeze(self, x, axis) -> ndx.Array:
        return NotImplemented

    # creation.py
    def full(self, shape, fill_value, dtype=None, device=None) -> ndx.Array:
        return NotImplemented

    def full_like(self, x, fill_value, dtype=None, device=None) -> ndx.Array:
        return NotImplemented

    def ones(
        self,
        shape,
        dtype: dtypes.CoreType | dtypes.StructType | None = None,
        device=None,
    ):
        return NotImplemented

    def ones_like(
        self, x, dtype: dtypes.StructType | dtypes.CoreType | None = None, device=None
    ):
        return NotImplemented

    def zeros(
        self,
        shape,
        dtype: dtypes.CoreType | dtypes.StructType | None = None,
        device=None,
    ):
        return NotImplemented

    def zeros_like(
        self, x, dtype: dtypes.CoreType | dtypes.StructType | None = None, device=None
    ):
        return NotImplemented

    def empty(self, shape, dtype=None, device=None) -> ndx.Array:
        return NotImplemented

    def empty_like(self, x, dtype=None, device=None) -> ndx.Array:
        return NotImplemented

    def arange(self, start, stop=None, step=None, dtype=None, device=None) -> ndx.Array:
        return NotImplemented

    def eye(self, n_rows, n_cols=None, k=0, dtype=None, device=None) -> ndx.Array:
        return NotImplemented

    def tril(self, x, k=0) -> ndx.Array:
        return NotImplemented

    def triu(self, x, k=0) -> ndx.Array:
        return NotImplemented

    def linspace(
        self, start, stop, num, *, dtype=None, device=None, endpoint=True
    ) -> ndx.Array:
        return NotImplemented

    # additional.py
    def fill_null(self, x, value) -> ndx.Array:
        return NotImplemented

    def shape(self, x) -> ndx.Array:
        return NotImplemented

    def make_nullable(self, x, null) -> ndx.Array:
        return NotImplemented

    def can_cast(self, from_, to) -> bool:
        return NotImplemented



# Copyright (c) QuantCo 2023-2024
# SPDX-License-Identifier: BSD-3-Clause

from collections.abc import Iterable, Sequence

import numpy as np

import ndonnx as ndx
import ndonnx._data_types as dtypes
import ndonnx._opset_extensions as opx

from ._interface import OperationsBlock
from ._utils import from_corearray


class UniformShapeOperations(OperationsBlock):
    """Provides implementation for shape/indexing operations that are generic across all
    data types where the array's shape is uniform across all of its constituent
    fields."""

    def shape(self, x):
        current = x
        while isinstance(current, ndx.Array):
            current = next(iter(current._fields.values()))
        return from_corearray(opx.shape(current))

    def take(self, x, indices, axis=None):
        if axis is None:
            axis = 0
        if isinstance(indices, ndx.Array):
            indices = indices._core()
        else:
            indices = opx.const(indices, dtype=dtypes.int64)
        return x._transmute(lambda corearray: opx.gather(corearray, indices, axis=axis))

    def broadcast_to(self, x, shape):
        shape = ndx.asarray(shape, dtype=dtypes.int64)._core()
        return x._transmute(lambda corearray: opx.expand(corearray, shape))

    def expand_dims(self, x, axis):
        return x._transmute(
            lambda corearray: opx.unsqueeze(
                corearray, axes=opx.const([axis], dtype=dtypes.int64)
            )
        )

    def flip(self, x, axis):
        if x.ndim == 0:
            return x.copy()

        if axis is None:
            axis = range(x.ndim)
        elif not isinstance(axis, Iterable):
            axis = [axis]

        axis = [x.ndim + ax if ax < 0 else ax for ax in axis]

        index = tuple(
            slice(None, None, None) if i not in axis else (slice(None, None, -1))
            for i in range(0, x.ndim)
        )

        return x[index]

    def permute_dims(self, x, axes):
        return x._transmute(lambda corearray: opx.transpose(corearray, perm=axes))

    def reshape(self, x, shape, *, copy=None):
        if (
            isinstance(shape, (list, tuple))
            and len(shape) == x.ndim == 1
            and shape[0] == -1
        ):
            return x.copy()
        else:
            x = ndx.asarray(x)
            shape = ndx.asarray(shape, dtype=dtypes.int64)._core()
            out = x._transmute(lambda corearray: opx.reshape(corearray, shape))
            if copy is False:
                return x._set(out)
            return out

    def roll(self, x, shift, axis):
        if not isinstance(shift, Sequence):
            shift = [shift]

        old_shape = x.shape

        if axis is None:
            x = ndx.reshape(x, [-1])
            axis = 0

        if not isinstance(axis, Sequence):
            axis = [axis]

        if len(shift) != len(axis):
            raise ValueError("shift and axis must have the same length")

        for sh, ax in zip(shift, axis):
            len_single = opx.gather(
                ndx.asarray(x.shape, dtype=dtypes.int64)._core(), opx.const(ax)
            )
            shift_single = opx.add(opx.const(-sh, dtype=dtypes.int64), len_single)
            # Find the needed element index and then gather from it
            range = opx.cast(
                opx.range(
                    opx.const(0, dtype=len_single.dtype),
                    len_single,
                    opx.const(1, dtype=len_single.dtype),
                ),
                to=dtypes.int64,
            )
            new_indices = opx.mod(opx.add(range, shift_single), len_single)
            x = ndx.take(x, from_corearray(new_indices), axis=ax)

        return ndx.reshape(x, old_shape)

    def squeeze(self, x, axis):
        axis = [axis] if not isinstance(axis, Sequence) else axis
        if len(axis) == 0:
            return x.copy()
        axes = opx.const(axis, dtype=dtypes.int64)
        return x._transmute(lambda corearray: opx.squeeze(corearray, axes=axes))

    # creation.py
    def full(self, shape, fill_value, dtype=None, device=None):
        fill_value = ndx.asarray(fill_value)
        if fill_value.ndim != 0:
            raise ValueError("fill_value must be a scalar")

        dtype = fill_value.dtype if dtype is None else dtype
        shape = ndx.asarray(shape, dtype=dtypes.int64)
        if shape.ndim == 0:
            shape = ndx.reshape(shape, [1])
        return fill_value._transmute(
            lambda corearray: opx.expand(corearray, shape._core())
        ).astype(dtype)

    def full_like(self, x, fill_value, dtype=None, device=None):
        fill_value = ndx.asarray(fill_value).astype(dtype or x.dtype)
        return ndx.broadcast_to(fill_value, ndx.asarray(x).shape)

    def where(self, condition, x, y):
        if x.dtype != y.dtype:
            target_dtype = ndx.result_type(x, y)
            x = ndx.astype(x, target_dtype)
            y = ndx.astype(y, target_dtype)
        if isinstance(condition.dtype, dtypes.Nullable) and not isinstance(
            x.dtype, (dtypes.Nullable, dtypes.CoreType)
        ):
            raise TypeError("where condition is nullable, but both outputs are not")
        if condition.to_numpy() is not None and condition.dtype == dtypes.bool:
            if (
                condition.to_numpy().size == 1
                and condition.to_numpy().ndim <= x.ndim
                and condition.to_numpy().item()
            ):
                return x.copy()
            elif (
                condition.to_numpy().size == 1
                and condition.to_numpy().ndim <= y.ndim
                and not condition.to_numpy().item()
            ):
                return y.copy()
        if x.dtype == y.dtype and x.to_numpy() is not None and y.to_numpy() is not None:
            if isinstance(x.to_numpy(), np.ma.MaskedArray):
                if np.ma.allequal(x.to_numpy(), y.to_numpy(), fill_value=False):
                    return ndx.asarray(
                        np.broadcast_arrays(x.to_numpy(), y.to_numpy())[0]
                    )
            else:
                cond = np.equal(x.to_numpy(), y.to_numpy())
                if isinstance(x.dtype, ndx.Floating):
                    cond |= np.isnan(x.to_numpy()) & np.isnan(y.to_numpy())
                if cond.all():
                    return ndx.asarray(
                        np.broadcast_arrays(x.to_numpy(), y.to_numpy())[0]
                    )
        condition_values = (
            condition.values if condition.dtype == ndx.nbool else condition
        )

        # generic layout operation, work over both arrays with same data type to apply the condition
        def where_dtype_agnostic(a: ndx.Array, b: ndx.Array):
            if a.dtype != b.dtype:
                raise ValueError("The dtype of both arrays must be the same")
            if a.dtype == dtypes.bool:
                return ndx.logical_xor(
                    ndx.logical_and(condition_values, a),
                    ndx.logical_and(~condition_values, b),
                )
            elif a.dtype in (dtypes.int8, dtypes.int16):
                return from_corearray(
                    opx.where(
                        condition_values._core(),
                        a.astype(dtypes.int32)._core(),
                        b.astype(dtypes.int32)._core(),
                    )
                ).astype(a.dtype)
            elif a.dtype in (dtypes.uint16, dtypes.uint32, dtypes.uint64):
                return from_corearray(
                    opx.where(
                        condition_values._core(),
                        a.astype(dtypes.int64)._core(),
                        b.astype(dtypes.int64)._core(),
                    )
                ).astype(a.dtype)
            elif isinstance(a.dtype, dtypes.CoreType):
                return from_corearray(
                    opx.where(
                        condition_values._core(),
                        a._core(),
                        b._core(),
                    )
                )
            else:
                fields = {}
                for name in a.dtype._fields():
                    fields[name] = where_dtype_agnostic(
                        getattr(a, name), getattr(b, name)
                    )
                return ndx.Array._from_fields(a.dtype, **fields)

        output = where_dtype_agnostic(x, y)
        if condition.dtype == ndx.nbool:
            # propagate null if present
            if not isinstance(output.dtype, dtypes.Nullable):
                output = ndx.additional.make_nullable(output, condition.null)
            else:
                output.null = output.null | condition.null
        return output

    def zeros_like(self, x, dtype=None, device=None):
        return ndx.zeros(x.shape, dtype=dtype or x.dtype, device=device)

    def ones_like(self, x, dtype=None, device=None):
        return ndx.ones(x.shape, dtype=dtype or x.dtype, device=device)



# Copyright (c) QuantCo 2023-2024
# SPDX-License-Identifier: BSD-3-Clause

from typing import Literal

import numpy as np
import pytest
import spox.opset.ai.onnx.v20 as op

import ndonnx as ndx
from ndonnx._propagation import eager_propagate


def test_add():
    candidate = ndx.asarray([1]) + 1
    np.testing.assert_array_equal(candidate.to_numpy(), np.array([2]), strict=True)


def indexing_model(mode: Literal["lazy", "constant"]):
    if mode == "constant":
        a = ndx.asarray([0, 1, 2, 3])
    else:
        a = ndx.array(
            shape=("N",),
            dtype=ndx.int64,
        )
    b = ndx.array(shape=("M",), dtype=ndx.int64)
    return ndx.build(
        {"a": a, "b": b} if mode == "lazy" else {"b": b}, {"y": a[1:3] * 2}
    )


def arithmetic_model(mode: Literal["lazy", "constant"]):
    if mode == "constant":
        a = ndx.asarray([0, 1, 2, 3])
    else:
        a = ndx.array(
            shape=("N",),
            dtype=ndx.int64,
        )
    b = ndx.array(
        shape=("N",),
        dtype=ndx.int64,
    )
    return ndx.build({"a": a, "b": b} if mode == "lazy" else {"b": b}, {"y": a * 2 + b})


def dynamic_masking_model(mode: Literal["lazy", "constant"]):
    if mode == "constant":
        a = ndx.asarray([0, 1, 2, 3], dtype=ndx.int64)
    else:
        a = ndx.array(
            shape=("N",),
            dtype=ndx.int64,
        )
    b = ndx.array(
        shape=("M",),
        dtype=ndx.int64,
    )
    c = ndx.array(
        shape=("N",),
        dtype=ndx.bool,
    )
    inputs = {"b": b, "c": c}
    if mode == "lazy":
        inputs["a"] = a
    return ndx.build(inputs, {"y": a[c] * 2 + b})


def constant_indexing_model(mode: Literal["lazy", "constant"]):
    if mode == "constant":
        a = ndx.asarray([0, 1, 2, 3], dtype=ndx.int64)
    else:
        a = ndx.array(
            shape=("N",),
            dtype=ndx.int64,
        )
    b = ndx.asarray([5, 7, 8, 8, 9, 9, 234], dtype=ndx.int64)
    idx = ndx.asarray([1, 3, 5, 0])
    result = a * b[idx]
    return ndx.build({"a": a} if mode == "lazy" else {}, {"y": result})


@pytest.mark.parametrize(
    "model_func, expected_operators_constant",
    [
        (
            indexing_model,
            {"Identity", "Constant"},
        ),
        (
            arithmetic_model,
            {"Identity", "Add", "Constant"},
        ),
        (
            # We should be blocked from any real constant propagation
            # here since we are indexing on an eager value.
            dynamic_masking_model,
            {
                "Identity",
                "Constant",
                "Mul",
                "Add",
                "Compress",
            },
        ),
        (
            constant_indexing_model,
            {"Constant", "Identity"},
        ),
    ],
)
def test_model_constant_folds(model_func, expected_operators_constant):
    operators_used_const = {node.op_type for node in model_func("constant").graph.node}
    assert operators_used_const == expected_operators_constant


@pytest.mark.parametrize(
    "a, b, op, expected_operators",
    [
        (
            ndx.array(shape=("N",), dtype=ndx.bool),
            ndx.array(shape=("N",), dtype=ndx.bool),
            ndx.logical_or,
            {"Identity", "Or"},
        ),
        (
            ndx.asarray([False], dtype=ndx.bool),
            ndx.array(shape=("N",), dtype=ndx.bool),
            ndx.logical_or,
            {"Identity"},
        ),
        (
            ndx.asarray([False, False], dtype=ndx.bool),
            ndx.array(shape=("N",), dtype=ndx.bool),
            ndx.logical_or,
            {"Identity", "Or", "Constant"},
        ),
        (
            ndx.array(shape=("N",), dtype=ndx.bool),
            ndx.array(shape=("N",), dtype=ndx.bool),
            ndx.logical_and,
            {"Identity", "And"},
        ),
        (
            ndx.asarray([True], dtype=ndx.bool),
            ndx.array(shape=("N",), dtype=ndx.bool),
            ndx.logical_and,
            {"Identity"},
        ),
        (
            ndx.asarray([True, True], dtype=ndx.bool),
            ndx.array(shape=("N",), dtype=ndx.bool),
            ndx.logical_and,
            {"Identity", "And", "Constant"},
        ),
    ],
)
def test_logical_folding(a, b, op, expected_operators):
    out = op(a, b)
    inputs = {}
    if a.to_numpy() is None:
        inputs["a"] = a
    if b.to_numpy() is None:
        inputs["b"] = b
    model_proto = ndx.build(inputs, {"out": out})
    operators_used_const = {node.op_type for node in model_proto.graph.node}
    assert operators_used_const == expected_operators


@pytest.mark.parametrize(
    "cond, x, y, expected_operators",
    [
        (
            ndx.asarray([True, True]),
            ndx.asarray(["a", "b"]),
            ndx.asarray(["c", "d"]),
            {"Identity", "Constant"},
        ),
        (
            ndx.array(shape=("N",), dtype=ndx.bool),
            ndx.asarray(["a", "b"]),
            ndx.asarray(["c", "d"]),
            {"Identity", "Constant", "Where"},
        ),
        (
            ndx.asarray([False, False]),
            ndx.asarray(["a", "b"]),
            ndx.asarray(["c", "d"]),
            {"Identity", "Constant"},
        ),
        (
            ndx.asarray([True, False]),
            ndx.asarray(["a", "b"]),
            ndx.asarray(["a", "b"]),
            {"Identity", "Constant"},
        ),
        (
            ndx.asarray([True, False]),
            ndx.asarray(["a", "a"]),
            ndx.asarray(["a"]),
            {"Identity", "Constant"},
        ),
        (
            ndx.array(shape=("N",), dtype=ndx.bool),
            ndx.asarray([np.nan]),
            ndx.asarray([np.nan]),
            {"Identity", "Constant"},
        ),
        (
            ndx.array(shape=("N",), dtype=ndx.bool),
            ndx.asarray(["a", "a"]),
            ndx.asarray(["a"]),
            {"Identity", "Constant"},
        ),
        (
            ndx.array(shape=("N",), dtype=ndx.bool),
            ndx.asarray(np.ma.masked_array(["a", "a"], mask=[True, False])),
            ndx.asarray(["a"]),
            {"Identity", "Constant", "And", "Xor", "Not", "Where"},
        ),
        (
            ndx.array(shape=("N",), dtype=ndx.bool),
            ndx.asarray(np.ma.masked_array(["a", "a"], mask=[True, False])),
            ndx.asarray(np.ma.masked_array(["a", "a"], mask=[True, False])),
            {"Identity", "Constant", "And", "Xor", "Not", "Where"},
        ),
    ],
)
def test_where_folding(cond, x, y, expected_operators):
    out = ndx.where(cond, x, y)
    inputs = {}
    if cond.to_numpy() is None:
        inputs["cond"] = cond
    if x.to_numpy() is None:
        inputs["x"] = x
    if y.to_numpy() is None:
        inputs["y"] = y
    model_proto = ndx.build(inputs, {"out": out})
    operators_used_const = {node.op_type for node in model_proto.graph.node}
    assert operators_used_const == expected_operators


def test_eager_propagation_nested_parameters():
    @eager_propagate
    def function(
        x: ndx.Array, mapping: dict[str, ndx.Array], seq: list[ndx.Array]
    ) -> tuple[ndx.Array, ndx.Array]:
        # do some spox stuff
        a = ndx.from_spox_var(op.sigmoid(mapping["a"].astype(ndx.float64).spox_var()))
        b = ndx.from_spox_var(
            op.regex_full_match(seq[0].spox_var(), pattern="^hello.*")
        )
        return (a + x) * mapping["b"], b

    x, y = function(
        ndx.asarray([1, 2, 3, 4]),
        {"a": ndx.asarray([1, -10, 120, 40]), "b": 10},
        [ndx.asarray(["a", "hello world", "world hello"])],
    )
    expected_x = np.asarray([17.310586, 20.000454, 40.0, 50.0])
    expected_y = np.asarray([False, True, False])

    np.testing.assert_allclose(x.to_numpy(), expected_x)
    np.testing.assert_array_equal(y.to_numpy(), expected_y)
